from collections import Counter
from dataclasses import dataclass
import matplotlib.pyplot as plt
from .workflow_tools import workflow, requires, WorkflowManager

from .lib_types import BarContainer, Collocate

STAGES = [
    "start", # Before any processing
    "collocates", # Manipulation of collocates 
    "count", # Operations dealing with the counts of the words appearing around the target word
    "end" # End of the workflow. After this stage the user must necessarily call result to continue with the analysis
]

@dataclass
class Ngrams:
    collocates: list[tuple[str, ...]]
    bigrams: list[Collocate]

class Collocations(WorkflowManager):
    """
    Extracts n-grams for a target word by a context window (both directed or undirected). Counts and plots the terms that co-occur with the target

    :param tokens: A list of tokenized words (e.g. from a text or corpus) in which to search for the target word and its collocates.
    
    """

    _start: list[str]
    _end: BarContainer

    def __init__(self, tokens: list[str] | None = None):
        """
        Initializes TargetCollocations with given attributes.
        """

        super().__init__()
        
        # Stages

        self._collocates: Ngrams
        self._count: Counter[str]

        self._set_workflow(input_arg=tokens)
    
    @requires(stages=["start"])
    @workflow(input_arg="tokens", input_process="_start", output_process="_collocates", next_stage="collocates")
    def extract_ngrams(self, *, tokens: list[str], target_word: str, window_size: tuple[int, int], directed: bool = False) -> Ngrams:
        """
        Extracts both the context words of the target collocation, returned as tuples whose lenght corresponds to the specified window_size, and the collocations of the target word, in either directed or undirected manner.

        :param tokens: A list of tokenized words (e.g. from a text or corpus) in which to search for the target word and its collocates.
        :param target_word: Target word whose collocation are to be identified.
        :param window_size: Context window represented as a tuple (left, right) of the number of words to be included to left and right of the target word.
        :param directed: If True, the position of collocates relative to the target word is considered (i.e., direction matters); if False, direction is ignored. Defaults to False. 
        """

        collocates: list[tuple[str, ...]] = []
        bigrams: list[Collocate] = []

        for index, word in enumerate(tokens):
            if word == target_word:
                start: int = max(0, index - window_size[0])
                end: int = min(len(tokens), index + window_size[1] + 1)

                left_context: list[str] = tokens[start:index]
                right_context: list[str] = tokens[index + 1:end]

                context: list[str] = left_context + right_context 

                collocates.append(tuple(context))
                
                if directed == True:
                    bigrams.extend([(word, ("L", target_word)) for word in left_context] + [(word, ("R", target_word)) for word in right_context])
                else:
                    bigrams.extend([(word, target_word) for word in context])
        
        return Ngrams(collocates, bigrams)
    
    @requires(stages=["collocates"])
    @workflow(input_arg="collocates", input_attrs={"collocates": "collocates"}, input_process="_collocates", output_process="_count", next_stage="count")
    def count_collocates(self, *, collocates: list[tuple[str, ...]]) -> Counter[str]:
        """
        Counts the collocates of the target word.

        :param collocates: A list of collocates to count. 
        """

        all_words: list[str] = [word for collocation in collocates for word in collocation]
        word_counts: Counter[str] = Counter(all_words)
        
        return word_counts

    @requires(stages=["count"])
    @workflow(input_arg="word_counts", input_process="_count", output_process="_end", next_stage="end")
    def plot(self, *, word_counts: Counter[str], size: int = 10, show: bool = True, path: str | None = None) -> BarContainer:
        """
        Plots the count of the collocates.

        :param word_counts: A Counter object with the counts of each word. Defaults to None.
        :param size: The number of the most common collocates to plot. Defaults to 10. 
        :param show: If True, shows the plot. Defaults to True.
        :param path: If provided, saves the plot in the specified path. Defaults to None.
        """

        counts: list[tuple[str, int]] = word_counts.most_common(size)
        
        words: tuple[str, ...]
        values: tuple[int, ...]
        words, values = zip(*counts)

        fig: BarContainer = plt.bar(words, values)
        plt.xlabel("Palabras")
        plt.ylabel("Cuentas")
        
        if path:
            plt.savefig(path)
        
        if show:
            plt.show()

        return fig