from sklearn.preprocessing import StandardScaler
from sklearn.decomposition import PCA, TruncatedSVD
from .workflow_tools import requires, workflow, WorkflowManager
from .matrix_base import MatrixRepresentation, accept_matrix_representation, matrix_to_dataframe

from .lib_types import ndarray, DataFrame

STAGES = [
    "start", # Before any transformation to the Co-Matrixis applied
    "embeddings", # Result of the embeddings
    "end" # Embeddings transformed to dataframe
]

class CountModels(WorkflowManager):
    """
    Generates word embeddings using dimensionality reduction techniques on a co-occurrence matrix, such as SVD and PCA.

    :param co_ocurrence_matrix: A Co-ocurrence matrix from which embeddings will be generated.
    """

    _start: ndarray | DataFrame
    _end: DataFrame

    def __init__(self, co_ocurrence_matrix: DataFrame | None = None):
        """
        Initializes CountModels with given attributes.
        """
        super().__init__()

        self._embeddings: MatrixRepresentation

        self._set_workflow(input_arg=co_ocurrence_matrix)

    @requires(stages=["start"])
    @workflow(input_arg="matrix", input_process="_start", output_process="_embeddings", next_stage="embeddings")
    @accept_matrix_representation()
    def scale_matrix(self, *, matrix: ndarray | DataFrame, **kwargs) -> MatrixRepresentation:
        """
        Scales the input matrix to have zero mean and unit variance for each feature.

        This method applies standardization using scikit-learn's StandardScaler, which transforms the data such that each colum (feature) has a mean of 0 and a standard deviation of 1.

        :param matrix: The input data to scale.
        :param kwargs: Additional keyword arguments to pass to sklearn's StandardScaler. For more information check: https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.StandardScaler.html

        This method supports different matrix forms due to decorator-based preprocessing:
            - matrix: ndarray or Dataframe
            - matrix representation: MatrixRepresentation
        """
        
        scaler: StandardScaler = StandardScaler(**kwargs)
        scaled_matrix: ndarray = scaler.fit_transform(matrix)

        return MatrixRepresentation(scaled_matrix, matrix.index if isinstance(matrix, DataFrame) else None, matrix.columns if isinstance(matrix, DataFrame) else None)

    @requires(stages=["embeddings"])
    @workflow(input_arg="matrix", input_process="_embeddings", output_process="_embeddings")
    @accept_matrix_representation(override=("columns", None))
    def svd_embeddings(self, *, matrix: ndarray | DataFrame, **kwargs) -> MatrixRepresentation:
        """
        Generates word embeddings using truncated Single Value Descomposition (SVD).

        :param matrix: A Co-occurrence matrix from which embeddings will be generated.
        :param kwargs: Additional keyword arguments to pass to sklearn's TruncatedSVD.
            Common options include:
                - n_components: Specifies the number of dimensions to reduce the Co-ocurrence matrix to. 
            For more information check: https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.TruncatedSVD.html

        This method supports different matrix forms due to decorator-based preprocessing:
            - matrix: ndarray or Dataframe
            - matrix representation: MatrixRepresentation
        """

        svd: TruncatedSVD = TruncatedSVD(**kwargs)
        embeddings: ndarray = svd.fit_transform(matrix)

        return MatrixRepresentation(embeddings, matrix.index if isinstance(matrix, DataFrame) else None, None)

    @requires(stages=["embeddings"])
    @workflow(input_arg="matrix", input_process="_embeddings", output_process="_embeddings")
    @accept_matrix_representation(override=("columns", None))
    def pca_embeddings(self, *, matrix: ndarray | DataFrame, **kwargs) -> MatrixRepresentation:
        """
        Generates word embeddings using Principal Component Analysis (PCA).

        :param matrix: A Co-occurrence matrix from which embeddings will be generated.
        :param kwargs: Additional keyword arguments to pass to sklearn's PCA.
            Common options include:
                - n_components: If an integer, specifies the number of dimensions to reduce the Co-ocurrence matrix to. If a float, the amount of variance to preserve during PCA.
            For more information check: https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.PCA.html

        This method supports different matrix forms due to decorator-based preprocessing:
            - matrix: ndarray or Dataframe
            - matrix representation: MatrixRepresentation
        """

        pca: PCA = PCA(**kwargs)
        embeddings: ndarray = pca.fit_transform(matrix)

        return MatrixRepresentation(embeddings, matrix.index if isinstance(matrix, DataFrame) else None, None)

    @requires(stages=["embeddings"])
    @workflow(input_arg="matrix", input_process="_embeddings", output_process="_end", next_stage="end")
    def to_dataframe(self, *, matrix: MatrixRepresentation, **kwargs) -> DataFrame:
        """
        Creates a dataframe from a matrix representation.

        :param matrix: An matrix represenation from which to create a dataframe.
        :param kwargs: Additional keyword arguments to pass to sklearn's pandas' DataFrame.
        """

        return matrix_to_dataframe(matrix=matrix, **kwargs)