from sklearn.neighbors import NearestNeighbors
import pandas as pd
import numpy as np
from sklearn.metrics.pairwise import cosine_similarity
from .workflow_tools import requires, workflow, WorkflowManager

from .lib_types import ndarray, Series, DataFrame

STAGES = [
    "start", # The embeddings array
    "end" # After a distance has been applied
]

class GeometricDistance(WorkflowManager):
    """
    Provides a set of methods to calculate the distance between the embeddings of words, such as Euclidean distance, Manhattan distance, Cosine similarity, Nearest neighbors, etc.
    """

    def __init__(self, embeddings: DataFrame | None = None):
        """
        Initializes GeometricDistance with given attributes.

        :param embeddings: A matrix of word embeddings.
        :param vocab: Sequence of words representing the vocabulary aligned with the embeddings.
        """

        super().__init__()

        self._set_workflow(input_arg=embeddings)

    @requires(stages=["start", "end"], multiple_calls=True)
    @workflow(input_arg="embeddings", input_process="_start", output_process="_end", next_stage="end") # It would be interesting if you could select a set of distances as then result return all of them for comparision. Or you could call different words and return an array wit the result of all of them.
    def euclidean_distance(self, *, embeddings: DataFrame, first_word: str, second_word: str) -> float:
        """
        Computes the Euclidean distance between the embeddings of two words.

        :param embeddings: A dataframe containing the word embeddings.
        :param first_word: The first word in the pair.
        :param second_word: The second word in the pair. 
        """

        first_word_vector: Series = embeddings.loc[first_word]
        second_word_vector: Series = embeddings.loc[second_word]

        return float(np.linalg.norm(first_word_vector - second_word_vector))

    @requires(stages=["start", "end"], multiple_calls=True)
    @workflow(input_arg="embeddings", input_process="_start", output_process="_end", next_stage="end")
    def manhattan_distance(self, *, embeddings: DataFrame, first_word: str, second_word: str) -> float:
        """
        Computes the Manhattan distance between the embeddings of two words.

        :param embeddings: A dataframe containing the word embeddings.
        :param first_word: The first word in the pair.
        :param second_word: The second word in the pair. 
        """

        first_word_vector: Series = embeddings.loc[first_word]
        second_word_vector: Series = embeddings.loc[second_word]

        return np.sum(np.abs(first_word_vector - second_word_vector))

    @requires(stages=["start", "end"], multiple_calls=True)
    @workflow(input_arg="embeddings", input_process="_start", output_process="_end", next_stage="end")
    def cos_similarity(self, *, embeddings: DataFrame, first_word: str, second_word: str) -> float:
        """
        Computes the cosine similarity between the embeddings of two words.

        :param embeddings: A dataframe containing the word embeddings.
        :param first_word: The first word in the pair.
        :param second_word: The second word in the pair. 
        """

        first_word_vector: ndarray = np.array(embeddings.loc[first_word]).reshape(1, -1)
        second_word_vector: ndarray = np.array(embeddings.loc[second_word]).reshape(1, -1)

        cos_sim: ndarray = cosine_similarity(first_word_vector, second_word_vector)

        return cos_sim[0][0]

    @requires(stages=["start", "end"], multiple_calls=True)
    @workflow(input_arg="embeddings", input_process="_start", output_process="_end", next_stage="end")
    def nearest_neighbors(self, *, embeddings: DataFrame, word: str, metric: str, n_words: int = 5, **kwargs) -> list[tuple[str, float]]:
        """
        Returns the top N most semantically similar words to a given target word, based on the specified distance or similarity metric.

        :param embeddings: A dataframe containing the word embeddings.
        :param word: The target word to find neighbors for.
        :param metric: The distance or similarity metric to use (e.g., 'cosine', 'euclidean').
        :param n_words: Number of nearest neighbors to return. Defaults to 5.
        :param kwargs: Additional keyword arguments to pass to sklearn's NearestNeighbors. For more information check: https://scikit-learn.org/stable/modules/generated/sklearn.neighbors.NearestNeighbors.html
        """

        neighbors: NearestNeighbors = NearestNeighbors(n_neighbors=n_words, algorithm="auto", metric=metric, **kwargs)
        neighbors.fit(embeddings.to_numpy())

        word_vector: Series = embeddings.loc[word]

        distances: ndarray
        indices: ndarray
        distances, indices = neighbors.kneighbors([word_vector], n_neighbors=n_words + 1)

        neighbor_tuples = zip(indices[0], distances[0])

        results: list[tuple[str, float]] = [(embeddings.index[index], 1 - distance) for index, distance in neighbor_tuples if embeddings.index[index] != word]

        return results