from dataclasses import dataclass, field
import pandas as pd
from .lib_types import Matrix, Labels, ndarray, DataFrame, csr_matrix, csc_matrix
from .workflow_tools import accepts_generic

from typing import Any, cast, Callable

@dataclass
class MatrixRepresentation:
    """
    Container for a matrix and its associated row and column labels, with optional metadata.

    :param matrix: The core matrix data, typically a NumPy array.
    :param rows: Optional row labels.
    :param columns: Optional column labels.
    :param meta: Optional dictionary to store additional metadata.
    """

    matrix: Matrix
    rows: Labels = None
    columns: Labels = None
    meta: dict[str, Any] = field(default_factory=dict)

def accept_matrix_representation(accepts: bool = True, custom_error_message: str = "", intercept: bool = True, meta: str | None = None, override: tuple[str, Any] | None = None) -> Callable:
    """
    Decorator that allows a method to accept a MatrixRepresentation object as input.

    Extracts the underlying matrix for processing and optionally reattaches metadata, labels, or captures non-matrix outputs into metadata.

    :param accepts: Whether to accept MatrixRepresentation inputs.
    :param custom_error_message: Optional error message if input type is not accepted.
    :param intercept: Whether to intercept the input and repackage the output.
    :param meta: Optional metadata key to store non-matrix outputs in the result.

    :return: A decorated method that supports MatrixRepresentation as input.
    """

    def is_matrix_representation(matrix: Any) -> bool:
        """
        If matrix is an instance of MatrixRepresentation, returns True. Else, returns False.

        :param matrix: A matrix to check its instance.
        """

        return True if isinstance(matrix, MatrixRepresentation) else False
        
    def intercept_matrix(self, input_value: MatrixRepresentation, method: Callable, *args, **kwargs) -> MatrixRepresentation:
        result: MatrixRepresentation | Any = method(self, *args, matrix=input_value.matrix, **kwargs)

        matrix: MatrixRepresentation

        if isinstance(result, MatrixRepresentation):
            matrix = result

            matrix.rows = input_value.rows
            matrix.columns = input_value.columns

            if override:
                attr, value = override
                setattr(matrix, attr, value) 
        else:
            if isinstance(result, ndarray):
                matrix = MatrixRepresentation(result, input_value.rows, input_value.columns)
            elif not isinstance(result, ndarray) and meta:
                matrix = MatrixRepresentation(input_value.matrix, input_value.rows, input_value.columns)
                matrix.meta[meta] = result

        return matrix
    
    return accepts_generic(
        type_checker=is_matrix_representation,
        input_arg="matrix",
        accepts=accepts,
        intercept=intercept,
        interceptor=lambda self, input_value, method, *args, **kwargs: intercept_matrix(self, input_value, method, *args, **kwargs),
        input_type=MatrixRepresentation,
        custom_error_message=custom_error_message
    )

def matrix_to_dataframe(matrix: MatrixRepresentation, **kwargs):
    """
    Converts a MatrixRepresentation to a pandas DataFrame.

    :param matrix: A MatrixRepresentation instance.
    :param kwargs: Additional keyword arguments to pass to sklearn's pandas' DataFrame.

    :return: A pandas DataFrame with corresponding data and labels.
    """

    if isinstance(matrix.matrix, csr_matrix) or isinstance(matrix.matrix, csc_matrix):
        matrix.matrix = matrix.matrix.toarray()

    return pd.DataFrame(
        matrix.matrix,
        index=matrix.rows,
        columns=matrix.columns,
        **kwargs
    )

def dataframe_to_matrix(dataframe: DataFrame):
    """
    Converts a pandas DataFrame to a MatrixRepresentation.

    :param dataframe: A pandas DataFrame.

    :return: A MatrixRepresentation with matrix data and index/column labels.
    """

    return MatrixRepresentation(
        matrix=dataframe.to_numpy(),
        rows=list(dataframe.index),
        columns=list(dataframe.columns)
    )