from .workflow_tools import WorkflowBuilder
from .pipeline_tools import PipelineBuilder
from dstk.lib_types import Language, DataFrame

from .text_processor import TextProcessor
from .text_matrix_builder import TextMatrixBuilder
from .weight_matrix import WeightMatrix
from .count_models import CountModels
from .geometric_distance import GeometricDistance
from .predict_models import PredictModels

def StandardModel(text: str, model: str | Language, window_size: int = 2, components: int = 100) -> DataFrame:
    """
    This pipeline generates word embeddings using the standard model as defined by (Lenci & Sahlgren 97). It preprocesses the text by removing stop words, lowering the words and segmenting the text using a context window. The co-occurrence matrix is weighted with PPMI and reduced with truncated SVD. 

    :param text: The text to extract the embeddings from.
    :param model: The spaCy NLP model to tokenize the text.
    :param window_size: The size of the context window to segment the text. Defaults to 2.
    :param components: The number of dimensions of the embeddings. Defaults to 100.
    """

    StandardTextWorkflow: WorkflowBuilder = WorkflowBuilder(
        work_class=TextProcessor, 
        method_representation={
            "set_model": {"model": model},
            "get_tokens": {},
            "remove_stop_words": {},
            "get_text": {"lemmatize": True},
            "to_lower": {},
            "corpus_by_context_window": {"window_size": window_size}
        }
    )

    StandardMatrix: WorkflowBuilder = WorkflowBuilder(
        work_class=TextMatrixBuilder,
        method_representation={
            "create_dtm": {},
            "create_co_occurrence_matrix": {},
            "to_dataframe": {}
        }
    )

    StandardWeightMatrix: WorkflowBuilder = WorkflowBuilder(
        work_class=WeightMatrix,
        method_representation={
            "pmi": {"positive": True}
        }
    )

    StandardCountModels: WorkflowBuilder = WorkflowBuilder(
        work_class=CountModels,
        method_representation={
            "scale_matrix": {},
            "svd_embeddings": {"n_components": components},
            "to_dataframe": {}
        }
    )

    Model: PipelineBuilder = PipelineBuilder([
        StandardTextWorkflow,
        StandardMatrix,
        StandardWeightMatrix,
        StandardCountModels
    ])

    return Model(text=text)

def SGNSModel(text: str, model: str | Language, path: str, **kwargs) -> PredictModels:
    """
    This pipeline generates word embeddings using Skip-Gram with Negative Sampling (SGNS) as defined by (Lenci & Sahlgren 162). It preprocesses the text by extracting the sentences, removing stop words and lowering them. The embeddings are extracted by using word2vec to do SGNS. Returns an instance of PredictModels.

    :param text: The text to extract the embeddings from.
    :param model: The spaCy NLP model to tokenize the text.
    :param path: The path to save the processed senteces.
    :param kwargs:  Additional keyword arguments to pass to gensim.models.Word2Vec. Common options include:
            - vector_size: Size of the word embedding vectors.
            - workers: Number of CPU cores to be used during the training process.
            - negative: Specifies how many "noise words" to sample for each positive example during training. Typical values range from 5 to 20. Higher values make training slower but can improve embedding quality.
            - window (int): Maximum distance between the current and predicted word.
            - min_count (int): Ignores all words with total frequency lower than this.

            For more information check: https://radimrehurek.com/gensim/models/word2vec.html
    """

    SGNSTextWorkflow: WorkflowBuilder = WorkflowBuilder(
        work_class=TextProcessor,
        method_representation={
            "set_model": {"model": model},
            "get_sentences": {},
            "remove_stop_words": {},
            "get_text": {"lemmatize": True},
            "to_lower": {},
            "join": {},
            "save_to_file": {"path": path}
        }
    )

    SGNSPredictWorkflow: WorkflowBuilder = WorkflowBuilder(
        work_class=PredictModels,
        method_representation={
            "word2vec": {
                "sg": 1,
                **kwargs
            }
        },
        result=False
    )

    Model: PipelineBuilder = PipelineBuilder([
        SGNSTextWorkflow,
        SGNSPredictWorkflow
    ])

    return Model(text=text)