import matplotlib.pyplot as plt
from sklearn.cluster import KMeans
from sklearn.metrics import silhouette_score
from kneed import KneeLocator
from umap import UMAP
from .workflow_tools import requires, workflow, WorkflowManager
from .matrix_base import MatrixRepresentation, accept_matrix_representation

from .lib_types import ndarray, PathCollection, Axes, Axes3D, Labels

STAGES = [
    "start", # The embeddings matrix representation
    "clusters", # The ideal number of clusters to be plotted
    "end" # The result of the plot
]

# Maybe return the interias and the highest score?
class PlotEmbeddings(WorkflowManager):
    """
    Provides a set of methods to visualize word embeddings in both 2D and 3D.

    :param embeddings: A matrix representation of word embeddings.
    """

    def __init__(self, embeddings: MatrixRepresentation | None = None):
        """
        Initializes PlotEmbeddings with given attributes.
        """

        super().__init__()

        # Stages

        self._clusters: tuple[int, float]

        self._set_workflow(input_arg=embeddings)

    @requires(stages=["start"])
    @workflow(input_arg="matrix", input_process="_start", output_process="_clusters", next_stage="clusters")
    @accept_matrix_representation(meta="n_clusters")
    def elbow_analysis(self, *, matrix: ndarray, max_k: int, show: bool = False, path: str | None = None) -> int:
        """
        Generates an Elbow plot to help determine the optimal number of clusters for the word embeddings. Returns the best number of clusters.

        :param matrix: An array to be clustered.
        :param max_k: The maximum number of clusters to evaluate when applying the Elbow method.
        :param show: If True, shows the plot. Defaults to False.
        :param path: If provided, saves the plot in the specified path. Defaults to None.

        This method supports different matrix forms due to decorator-based preprocessing:
            - matrix: ndarray
            - matrix representation: MatrixRepresentation
        """
        plt.clf() 
        plt.close("all")

        means: list[int] = []
        inertias: list[float] = []

        for k in range(1, max_k):
            kmeans: KMeans = KMeans(n_clusters=k, random_state=42)
            kmeans.fit(matrix)

            means.append(k)
            inertias.append(kmeans.inertia_)

        elbow: KneeLocator = KneeLocator(means, inertias, curve="convex", direction="decreasing")

        print(f"The best cluster is {elbow.knee} with an inertia of {elbow.knee_y}")

        elbow.plot_knee()

        if path:
            plt.savefig(path)

        if show:
            plt.show()

        return elbow.knee

    @requires(stages=["start"])
    @workflow(input_arg="matrix", input_process="_start", output_process="_clusters", next_stage="clusters")
    @accept_matrix_representation(meta="n_clusters")
    def extract_silhouette_score(self, *, matrix: ndarray, max_k: int, show: bool = False, path: str | None = None, **kwargs) -> int:
        """
        Extracts and plots the Silhouette score to help determine the optimal number of clusters for the word embeddings. Returns the best number of clusters based on the highest score.

        :param matrix: An array to be clustered.
        :param max_k: The maximum number of clusters to evaluate when computing the Silhouette score.
        :param show: If True, shows the plot. Defaults to False.
        :param path: If provided, saves the plot in the specified path. Defaults to None.
        :param kwargs: Additional keyword arguments to pass to sklearn.metrics silhouette_score. For more information check: https://scikit-learn.org/stable/modules/generated/sklearn.metrics.silhouette_score.html

        This method supports different matrix forms due to decorator-based preprocessing:
            - matrix: ndarray
            - matrix representation: MatrixRepresentation
        """
        plt.clf() 
        plt.close("all")

        sil_scores: list[tuple[int, float]] = []

        for k in range(2, max_k):
            kmeans: KMeans = KMeans(n_clusters=k, random_state=42)
            kmeans.fit(matrix)
            sil_score: float = silhouette_score(matrix, kmeans.labels_)
            sil_scores.append((k, sil_score), **kwargs)

        highest_score: tuple[int, float] = max(sil_scores, key=lambda tup: tup[1])
        print(f"The best cluster is {highest_score[0]} with a Silhouette score of {highest_score[1]}")

        clusters, scores = zip(*sil_scores)
        
        plt.plot(clusters, scores, 'o-')

        plt.xlabel("Number of Clusters")
        plt.ylabel("Silhouette Scores")
        plt.title("Silhouette Score Plot")
        plt.grid(True)

        if path:
            plt.savefig(path)

        if show:
            plt.show()

        return highest_score[0]

    @requires(stages=["clusters"])
    @workflow(input_arg="embeddings", input_attrs={"embeddings": None, "n_clusters": {"meta": "n_clusters"}}, input_process="_clusters", output_process="_end", next_stage="end") 
    def plot_embeddings_2D(self, *, embeddings: MatrixRepresentation, n_clusters: int = 1, alpha: float = 0.8, font_size: int = 8, grid: bool = True, show: bool = True, path: str | None = None, umap_neighbors: int = 15, umap_metric: str = "cosine", umap_dist: float = 0.1) -> PathCollection:
        """
        Generates a 2D plot of the word embedddings using UMAP for dimensionality reduction.

        :param embeddings: A matrix representation of word embeddings.
        :param n_clusters: The number of clusters to form when grouping the word embeddings. Defaluts to 1.
        :param alpha: The transparency level (alpha) of the plotted dots, between 0 (fully transparent) and 1 (fully opaque). Defaults to 0.8.
        :param font_size: Specifies the font size for the labels of the plotted dots. Defaults to 8.
        :param grid: If True, displays a grid on the plot; if False, hides it. Defaults to True.
        :param show: If True, shows the plot. Defaults to True.
        :param path: If provided, saves the plot in the specified path. Defaults to None.
        :param umap_neighbors: Controls how UMAP balances local versus global structure. Higher values consider a broader context when reducing dimensions. Defaults to 15.
        :param umap_metric: The distance metric UMAP uses to assess similarity between words (e.g., "cosine", "euclidean"). Defaults to "cosine", which is common for word embeddings.
        :param umap_dist: Controls how tightly UMAP packs points together. Lower values keep similar words closer in the 2D space. Defaults to 0.1.

        :return: A matplotlib PathCollection object representing the plotted points.
        """

        plt.clf()
        plt.close("all")

        embeddings_matrix: ndarray = embeddings.matrix
        labels: Labels = embeddings.rows

        reducer: UMAP = UMAP(n_components=2, n_neighbors=umap_neighbors, min_dist=umap_dist, metric=umap_metric)
        umap_embeddings: ndarray = reducer.fit_transform(embeddings_matrix)

        kmeans: KMeans = KMeans(n_clusters=n_clusters, random_state=42)
        clusters: ndarray = kmeans.fit_predict(umap_embeddings)
        scatter: PathCollection = plt.scatter(umap_embeddings[:, 0], umap_embeddings[:, 1], c=clusters, cmap='Spectral', alpha=alpha)

        if labels is not None:
            for index, label in enumerate(labels):
                plt.annotate(label, (umap_embeddings[index, 0], umap_embeddings[index, 1]), fontsize=font_size, alpha=0.6)
        else:
            raise ValueError("Rows of embeddings are None. Make sure your matrix contains labeled rows.")

        plt.xlabel("Axis 1")
        plt.ylabel("Axis 2")
        plt.title("Projection of word embeddings")
        plt.colorbar(scatter, label="Cluster")
        plt.grid(grid)
        plt.tight_layout()

        if path:
            plt.savefig(path)

        if show:
            plt.show()
        
        return scatter

    @requires(stages=["clusters"])
    @workflow(input_arg="embeddings", input_attrs={"embeddings": None, "n_clusters": {"meta": "n_clusters"}}, input_process="_clusters", output_process="_end", next_stage="end") 
    def plot_embeddings_3D(self, *, embeddings: MatrixRepresentation, n_clusters: int = 1, alpha: float = 0.8, font_size: int = 8, grid: bool = True, show: bool = True, path: str | None = None, umap_neighbors: int = 15, umap_metric: str = "cosine", umap_dist: float = 0.1) -> PathCollection:
        """
        Generates a 3D plot of the word embedddings.

        :param embeddings: A matrix representation of word embeddings.
        :param n_clusters: The number of clusters to form when grouping the word embeddings. Defaults to 1.
        :param alpha: The transparency level (alpha) of the plotted dots, between 0 (fully transparent) and 1 (fully opaque). Defaults to 0.8.
        :param font_size: Specifies the font size for the labels of the plotted dots. Defaults to 8.
        :param grid: If True, displays a grid on the plot; if False, hides it. Defaults to True.
        :param show: If True, shows the plot. Defaults to True.
        :param path: If provided, saves the plot in the specified path. Defaults to None.
        :param umap_neighbors: Controls how UMAP balances local versus global structure. Higher values consider a broader context when reducing dimensions. Defaults to 15.
        :param umap_metric: The distance metric UMAP uses to assess similarity between words (e.g., "cosine", "euclidean"). Defaults to "cosine", which is common for word embeddings.
        :param umap_dist: Controls how tightly UMAP packs points together. Lower values keep similar words closer in the 2D space. Defaults to 0.1.

        :return: A matplotlib PathCollection object representing the plotted points.
        """
        plt.clf()
        plt.close("all")

        embeddings_matrix: ndarray = embeddings.matrix
        labels: Labels = embeddings.rows

        reducer: UMAP = UMAP(n_components=3, n_neighbors=umap_neighbors, min_dist=umap_dist, metric=umap_metric)
        umap_embeddings: ndarray = reducer.fit_transform(embeddings_matrix)

        kmeans: KMeans = KMeans(n_clusters=n_clusters, random_state=42)
        clusters = kmeans.fit_predict(umap_embeddings)

        ax: Axes3D = plt.axes(projection="3d")

        scatter: PathCollection = ax.scatter(umap_embeddings
        [:, 0], umap_embeddings[:, 1], umap_embeddings[:, 2], c=clusters, cmap="Spectral", alpha=alpha)

        if labels is not None:
            for index, label in enumerate(labels):
                ax.text(umap_embeddings[index, 0], umap_embeddings[index, 1], umap_embeddings[index, 2], label, size=font_size, alpha=0.6)
        else:
            raise ValueError("Rows of embeddings are None. Make sure your matrix contains labeled rows.")

        ax.set_xlabel("Axis 1")
        ax.set_ylabel("Axis 2")
        ax.set_zlabel("Axis 3")
        ax.set_title("3D projection of word embeddings")

        plt.colorbar(scatter, label="Cluster")
        plt.grid(grid)
        plt.tight_layout()
  
        if path:
            plt.savefig(path)

        if show:
            plt.show()
        
        return scatter