import pandas as pd
from sklearn.feature_extraction.text import CountVectorizer
import numpy as np
from .workflow_tools import requires, workflow, WorkflowManager
from .matrix_base import MatrixRepresentation, matrix_to_dataframe

from .lib_types import csr_matrix, ndarray, DataFrame, csc_matrix, Matrix, Labels

STAGES = [
    "start", # Before any processing
    "matrix_operations" # Matrix operations and transformations
    "end" # Dataframe operations
]

class TextMatrixBuilder(WorkflowManager):
    """
    Creates a Document Term Matrix, a Co-ocurrence Matrix, and dataframes from them.

    :param corpus: A list of sentences or collocations from which to build a matrix.
    """

    _start: list[str]
    _end: DataFrame

    def __init__(self, corpus: list[str] | None = None):
        """
        Initializes TextMatrixBuilder with given attributes.
        """

        super().__init__()

        self._matrix_operations: MatrixRepresentation
        
        self._set_workflow(input_arg=corpus)

    @requires(stages=["start"])
    @workflow(input_arg="corpus", input_process="_start", output_process="_matrix_operations", next_stage="matrix_operations")
    def create_dtm(self, *, corpus: list[str], **kwargs) -> MatrixRepresentation:
        """
        Creates Document Term Matrix (DTM).

        :param corpus: A list of sentences or collocations from which to build a matrix.
        :param kwargs: Additional keyword arguments to pass to sklearn's CountVectorizer.
            Common options include:
                - stop_words: If provided, a list of stopwords to remove from the corpus.
                - ngram_range: A tuple (min_n, max_n) specifying the range of n-grams to consider.
            For more information check: https://scikit-learn.org/stable/modules/generated/sklearn.feature_extraction.text.CountVectorizer.html
        """

        vectorizer: CountVectorizer = CountVectorizer(**kwargs)

        dtm: csr_matrix = vectorizer.fit_transform(corpus)

        return MatrixRepresentation(
            matrix=dtm,
            rows=np.array(corpus), 
            columns=vectorizer.get_feature_names_out()
        )

    @requires(stages=["matrix_operations"])
    @workflow(input_arg="matrix", input_attrs={"matrix": "matrix", "rows": "columns", "columns": "columns"},  input_process="_matrix_operations", output_process="_matrix_operations")
    def create_co_occurrence_matrix(self, *, matrix: csr_matrix, rows: Labels = None, columns: Labels = None) -> csc_matrix:
        """
        Creates a Co-occurrence matrix.

        :param matrix: A Document Term Matrix (DTM) from which o build a Co-occurrence matrix.
        """

        co_matrix = matrix.T @ matrix

        return MatrixRepresentation(
            matrix=co_matrix,
            rows=rows,
            columns=columns
        )

    @requires(stages=["matrix_operations"])
    @workflow(input_arg="matrix", input_process="_matrix_operations", output_process="_end", next_stage="end")
    def to_dataframe(self, *, matrix: MatrixRepresentation, **kwargs) -> DataFrame:
        """
        Creates a dataframe from a matrix representation.

        :param matrix: A matrix representation from which to create a dataframe.
        :param kwargs: Additional keyword arguments to pass to sklearn's pandas' DataFrame.
        """
        
        return matrix_to_dataframe(matrix=matrix, **kwargs)