import pandas as pd
import numpy as np
from sklearn.feature_extraction.text import TfidfTransformer
from .workflow_tools import requires, workflow, WorkflowManager

from .lib_types import DataFrame, ndarray, Series, csr_matrix

STAGES = [
    "start", # Before any processing
    "end" # Weighted dataframe
]

class WeightMatrix(WorkflowManager):
    """
    Provides a set of methods to weight a Co-ocurrence matrix.

    :param dataframe: A Co-occurrence matrix to be weighted.
    """

    _start: DataFrame
    _end: DataFrame

    def __init__(self, dataframe: DataFrame | None = None):
        """
        Initializes WeightMatrix with given attributes.
        """

        super().__init__()

        self._set_workflow(input_arg=dataframe)

    @requires(stages=["start"])
    @workflow(input_arg="dataframe", input_process="_start", output_process="_end", next_stage="end")
    def pmi(self, *, dataframe: DataFrame, positive: bool = False) -> DataFrame:
        """
        Weights a Co-occurrence matrix by PMI or PPMI.
        
        :param dataframe: A Co-occurrence matrix to be weighted.
        :param positive: If True, weights the Co-ocurrence matrix by PPMI. If False, weighths it by PMI. Defaults to False.
        """

        df: DataFrame = dataframe

        col_totals: Series = df.sum(axis=0)
        total: float = col_totals.sum()
        row_totals: Series = df.sum(axis=1)
        expected: ndarray = np.outer(row_totals, col_totals) / total
        df = df / expected
        # Silence distracting warnings about log(0):
        with np.errstate(divide='ignore'):
            df = np.log(df)
        df[np.isinf(df)] = 0.0  # log(0) = 0
        if positive:
            df[df < 0] = 0.0

        return df

    @requires(stages=["start"])
    @workflow(input_arg="dataframe", input_process="_start", output_process="_end", next_stage="end")
    def tf_idf(self, *, dataframe: DataFrame, **kwargs) -> DataFrame:
        """
        Weights a Co-occurrence matrix by Tf-idf.
        
        :param dataframe: A Co-occurrence matrix to be weighted.
        """

        transformer: TfidfTransformer = TfidfTransformer(**kwargs)
        tf_idf_matrix: csr_matrix = transformer.fit_transform(dataframe)

        return pd.DataFrame(tf_idf_matrix.toarray(), index=dataframe.index, columns=dataframe.columns)
        